#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import numpy as np

from climetlab import load_source
from climetlab.utils.dates import parse_date, to_datetime, to_datetime_list


def test_to_datetime():
    pydate = datetime.datetime(2016, 1, 1)

    assert to_datetime(np.datetime64("2016-01-01")) == pydate

    assert to_datetime(np.datetime64("2016-01-01 00:00:00")) == pydate

    assert to_datetime(datetime.date(2016, 1, 1)) == pydate

    assert to_datetime("2016-01-01") == pydate

    assert to_datetime("2016-01-01T00:00:00") == pydate


def test_parse_date():
    assert parse_date("1851-06-25T00:00") == datetime.datetime(1851, 6, 25)
    assert parse_date("1851-06-25T06:00") == datetime.datetime(1851, 6, 25, 6)
    assert parse_date("1851-06-25") == datetime.datetime(1851, 6, 25)

    assert parse_date("18510625") == datetime.datetime(1851, 6, 25)
    assert parse_date(18510625) == datetime.datetime(1851, 6, 25)

    assert parse_date("1851-06-25 06:00:00") == datetime.datetime(1851, 6, 25, 6)
    assert parse_date("1851-06-25T06:00:00") == datetime.datetime(1851, 6, 25, 6)
    assert parse_date("1851-06-25T06:00:00Z") == datetime.datetime(
        1851, 6, 25, 6, tzinfo=datetime.timezone.utc
    )

    assert parse_date(-2) == parse_date(0) - datetime.timedelta(days=2)


def test_to_datetimes_list():
    assert to_datetime_list("20000101/to/20000103") == [
        datetime.datetime(2000, 1, 1),
        datetime.datetime(2000, 1, 2),
        datetime.datetime(2000, 1, 3),
    ]
    assert to_datetime_list("2000-01-01/to/2000-01-03") == [
        datetime.datetime(2000, 1, 1),
        datetime.datetime(2000, 1, 2),
        datetime.datetime(2000, 1, 3),
    ]
    assert to_datetime_list("2000-01-01/to/2000-01-10/by/3") == [
        datetime.datetime(2000, 1, 1),
        datetime.datetime(2000, 1, 4),
        datetime.datetime(2000, 1, 7),
        datetime.datetime(2000, 1, 10),
    ]
    assert to_datetime_list((20000101, "to", 20000103)) == [
        datetime.datetime(2000, 1, 1),
        datetime.datetime(2000, 1, 2),
        datetime.datetime(2000, 1, 3),
    ]

    assert to_datetime_list(("2000-01-01", "to", "2000-01-10", "by", "3")) == [
        datetime.datetime(2000, 1, 1),
        datetime.datetime(2000, 1, 4),
        datetime.datetime(2000, 1, 7),
        datetime.datetime(2000, 1, 10),
    ]

    assert len(to_datetime_list((-10, "to", -1))) == 10


def test_to_datetimes_list_grib():
    source = load_source("file", "docs/examples/test.grib")
    for s in source:
        assert to_datetime_list(s) == [datetime.datetime(2020, 5, 13, 12, 0)]


if __name__ == "__main__":
    # test_to_datetime()
    to_datetime_list()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import itertools
from collections import defaultdict

from . import DataSource


class MultiSource(DataSource):
    def __init__(self, *sources):
        if len(sources) == 1 and isinstance(sources[0], list):
            sources = sources[0]

        self.sources = sources
        self._lengths = [None] * len(sources)

    # def not_yet_mutate(self):
    #     if len(self.sources) == 1:
    #         return self.sources[0].mutate()

    #     t = type(self.sources[0])
    #     if all(type(s) == t for s in self.sources):
    #         return t.multi_merge(self.sources).mutate()

    #     return self

    def _set_dataset(self, dataset):
        super()._set_dataset(dataset)
        for s in self.sources:
            s._set_dataset(dataset)

    def __iter__(self):
        return itertools.chain(*self.sources)

    def __getitem__(self, n):

        if n < 0:
            n = len(self) + n

        i = 0
        while n >= self._length(i):
            n -= self._length(i)
            i += 1
        return self.sources[i][n]

    def sel(self, *args, **kwargs):
        raise NotImplementedError
        # new_sources = []
        # for s in self.sources:
        #     new = s.sel(*args, **kwargs)
        #     return new_sources.append(new)
        # merged = self.__class__(self, sources=new_sources)
        # return merged

    def __len__(self):
        return sum(self._length(i) for i, _ in enumerate(self.sources))

    def _length(self, i):
        if self._lengths[i] is None:
            self._lengths[i] = len(self.sources[i])
        return self._lengths[i]

    def to_xarray(self):
        import xarray as xr

        arrays = [s.to_xarray() for s in self.sources]

        # Get values of scalar coordinates
        values = defaultdict(set)
        for a in arrays:
            for v in a.coords:
                if len(a[v].shape) == 0:
                    vals = a[v].values
                    # assert len(vals) == 1, (v, vals)
                    values[v].add(float(vals))

        # for a in arrays:
        #     print("++++ ======")
        #     for v in a.variables:
        #         print(v, [x for x in a[v].dims])
        #     print("++++ ======")
        #     print()

        # Promote scalar coordinates
        promote = [name for name, count in values.items() if len(count) > 1]

        if promote:
            dims = dict(zip(promote, [1] * len(promote)))
            arrays = [a.expand_dims(dims) for a in arrays]

        # return xr.concat(s.to_xarray() for s in self.sources)
        return xr.merge(arrays)


source = MultiSource



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import yaml


def test_yaml():
    print(yaml.__version__)
    yfile = os.path.dirname(__file__) + "/example.yaml"
    print(yfile)
    with open(yfile) as f:
        s = yaml.load(f.read(), Loader=yaml.SafeLoader)
        print(s)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import yaml

from climetlab.core.caching import temp_file
from climetlab.core.ipython import SVG, Image
from climetlab.core.metadata import annotation
from climetlab.utils.bbox import BoundingBox

from .actions import mcoast, mgrib, minput, mmap, mnetcdf, mtable, mtext, output, plot
from .apply import apply

LOG = logging.getLogger(__name__)


class Layer:
    def __init__(self, data):
        self._data = data
        self._style = data.default_style

    def add_action(self, actions):
        if self._data:
            actions.append(self._data)
        if self._style:
            actions.append(self._style)

    def style(self, style):
        self._style = style

    def update(self, action, value):
        return self._style.update(action, value)


class WMSLayer:
    def __init__(self, actions, bounding_box, temp_files):
        self.actions = actions
        self.bounding_box = bounding_box
        self.temp_files = temp_files


class Driver:
    """TODO: Docscting"""

    def __init__(self, options):

        self._options = options

        self._projection = None
        self._background = None
        self._foreground = None

        self._layers = []
        self._width_cm = 10.0
        self._height_cm = 10.0

        self._page_ratio = 1.0

        self.background(True)
        self.foreground(True)

        self._grid = None
        self._rivers = None
        self._cities = None
        self._borders = None

        self._legend = None
        self._title = None

        self._bounding_box = None
        self._tmp = []

    def temporary_file(self, extension: str = ".tmp") -> str:
        """Return a temporary file name that will be deleted once the plot is produced.

        :param extension: File name extension, defaults to ".tmp"
        :type extension: str, optional
        :return: Temporary file name.
        :rtype: str
        """

        if self._options("dump_yaml", False):
            return temp_file(extension).path

        self._tmp.append(temp_file(extension))
        return self._tmp[-1].path

    def bounding_box(self, north: float, west: float, south: float, east: float):

        bbox = BoundingBox(north=north, west=west, south=south, east=east)
        if self._bounding_box is None:
            self._bounding_box = bbox
        else:
            self._bounding_box = self._bounding_box.merge(bbox)

    def _push_layer(self, data):
        self._layers.append(Layer(data))

    def plot_grib(self, path: str, offset: int):

        self._push_layer(
            mgrib(
                grib_input_file_name=path,
                grib_file_address_mode="byte_offset",
                grib_field_position=int(offset),
            )
        )

    def plot_netcdf(self, path: str, variable: str, dimensions: dict = None):

        if dimensions is None:
            dimensions = {}

        dimension_setting = ["%s:%s" % (k, v) for k, v in dimensions.items()]

        if dimension_setting:
            params = dict(
                netcdf_filename=path,
                netcdf_value_variable=variable,
                netcdf_dimension_setting=dimension_setting,
                netcdf_dimension_setting_method="index",
            )
        else:
            params = dict(netcdf_filename=path, netcdf_value_variable=variable)

        self._push_layer(mnetcdf(**params))

    def plot_numpy(
        self,
        data,
        metadata: dict = None,
    ):
        if metadata is None:
            metadata = {}

        # TODO: issue warnings
        north = metadata.get("north", 90)
        west = metadata.get("west", 0)
        south_north_increment = metadata.get("south_north_increment")
        west_east_increment = metadata.get("west_east_increment")

        if south_north_increment is None:
            south_north_increment = (north - metadata.get("south", -90)) / (
                data.shape[-2] - 1
            )

        if west_east_increment is None:
            west_east_increment = (metadata.get("east", 360) - west) / (
                data.shape[-1] - 1
            )

        # TODO: remove me when Magics supports full json
        def tidy(x):
            r = {}
            for k, v in x.items():
                if v is None or isinstance(v, (int, float, str, bool)):
                    r[k] = v
            return r

        self._push_layer(
            minput(
                input_field=data,
                input_field_initial_latitude=float(north),
                input_field_latitude_step=-float(south_north_increment),
                input_field_initial_longitude=float(west),
                input_field_longitude_step=float(west_east_increment),
                input_metadata=tidy(metadata),
            )
        )

    def plot_xarray(self, ds, variable: str, dimensions: dict = None):
        tmp = self.temporary_file(".nc")
        field = ds[variable].isel({} if dimensions is None else dimensions)
        field.to_netcdf(tmp)
        self.plot_netcdf(tmp, variable, {} if dimensions is None else dimensions)

    def plot_csv(self, path: str, variable: str):
        self._push_layer(
            mtable(
                table_filename=path,
                table_latitude_variable="1",
                table_longitude_variable="2",
                table_value_variable="3",
                table_header_row=0,
                table_variable_identifier_type="index",
            )
        )
        self.style("default-style-observations")

    def plot_pandas(self, frame, latitude: str, longitude: str, variable: str):
        tmp = self.temporary_file(".csv")
        frame[[latitude, longitude, variable]].to_csv(tmp, header=False, index=False)
        self.plot_csv(tmp, variable)

        style = annotation(frame).get("style")
        if style is not None:
            self.style(style)

    def background(self, background):
        self._background = apply(
            value=background,
            collection="layers",
            target=self._background,
            default="default-background",
            options=self._options,
        )

    def foreground(self, foreground):
        self._foreground = apply(
            value=foreground,
            collection="layers",
            target=self._foreground,
            default="default-foreground",
            options=self._options,
        )

    def projection(self, projection):
        self._projection = apply(
            value=projection,
            collection="projections",
            target=self._projection,
            default=None,
            options=self._options,
        )

    def style(self, style):
        if len(self._layers) > 0:
            last_layer = self._layers[-1]
            last_layer.style(
                apply(
                    value=style,
                    target=last_layer,
                    collection="styles",
                    default=None,
                    options=self._options,
                )
            )
        else:
            raise Exception("No current data layer: cannot set style '%r'" % (style,))

    def apply_options(self, options):
        if options.provided("style"):
            self.style(options["style"])

        if options.provided("bounding_box"):
            bbox = options["bounding_box"]
            if isinstance(bbox, (list, tuple)):
                self.bounding_box(
                    north=bbox[0], west=bbox[1], south=bbox[2], east=bbox[3]
                )
            else:
                self.bounding_box(
                    north=bbox.north, west=bbox.west, south=bbox.south, east=bbox.east
                )

    def option(self, name, default=None):
        return self._options(name, default)

    def finalise(self):
        self.apply_options(self._options)

        if self._options.provided("background"):
            self.background(self._options["background"])

        if self._options.provided("foreground"):
            self.foreground(self._options["foreground"])

        if self._options.provided("projection"):
            self.projection(self._options["projection"])

        if self._options("grid", False):
            self._grid = mcoast(map_grid=True, map_coastline=False)

        if self._options("borders", False):
            self._borders = mcoast(
                map_boundaries=True,
                map_grid=False,
                map_coastline=False,
                map_label=False,
            )

        if self._options("rivers", False):
            self._rivers = mcoast(
                map_rivers=True, map_grid=False, map_coastline=False, map_label=False
            )

        if self._options("cities", False):
            self._cities = mcoast(
                map_cities=True, map_label=False, map_grid=False, map_coastline=False
            )

    def show(self):

        self.finalise()

        title = self._options("title", None)
        width = self._options("width", 680)
        frame = self._options("frame", False)

        path = self._options(
            "path", self.temporary_file("." + self._options("format", "png"))
        )

        if self._projection is None:
            # TODO: select best projection based on bbox
            self._projection = mmap(subpage_map_projection="cylindrical")

        if self._bounding_box is not None:
            bbox = self._bounding_box.add_margins(self._options("margins", 0))
            self._projection = apply(
                value={
                    "=subpage_upper_right_longitude": bbox.east,
                    "=subpage_upper_right_latitude": bbox.north,
                    "=subpage_lower_left_latitude": bbox.south,
                    "=subpage_lower_left_longitude": bbox.west,
                },
                target=self._projection,
                action=mmap,
                default=None,
                options=self._options,
            )

        self._page_ratio = self._projection.page_ratio()
        if self._page_ratio <= 0:
            self._page_ratio = 1.0

        _title_height_cm = 0
        if title:
            _title_height_cm = 0.7
            if title is True:
                # Automatic title
                self._title = mtext()
            else:
                self._title = mtext(
                    text_lines=[str(title)],
                    # text_justification='center',
                    # text_font_size=0.6,
                    # text_mode="positional",
                    # text_box_x_position=5.00,
                    # text_box_y_position=18.50,
                    # text_colour='charcoal'
                )

        page = output(
            output_file=path,
            page_x_length=self._width_cm,
            page_y_length=self._height_cm * self._page_ratio,
            super_page_x_length=self._width_cm,
            super_page_y_length=self._height_cm * self._page_ratio + _title_height_cm,
            subpage_x_length=self._width_cm,
            subpage_y_length=self._height_cm * self._page_ratio,
            subpage_x_position=0.0,
            subpage_y_position=0.0,
            output_width=width,
            page_frame=frame,
            page_id_line=False,
        )

        # TODO
        self._options("update", False)
        self._options("update_foreground", False)

        args = [page] + self.macro()

        if self._options("dump_python", False):
            print(args)

        dump = self._options("dump_yaml", False)
        if dump:
            if isinstance(dump, str):
                with open(dump, "w") as f:
                    print(
                        yaml.dump(
                            dict(plot=[a.to_yaml() for a in args]),
                            default_flow_style=False,
                        ),
                        file=f,
                    )
            else:
                print(
                    yaml.dump(
                        dict(plot=[a.to_yaml() for a in args]), default_flow_style=False
                    )
                )

        self._options.check_unused()

        try:
            plot(*args)
        except Exception:
            LOG.error("Error executing: %r", args, exc_info=True)
            raise

        if path.endswith(".svg"):
            Display = SVG  # noqa: N806
        elif path.endswith(".pdf"):
            return path
        else:
            Display = Image  # noqa: N806

        return Display(path, metadata=dict(width=width))

    def macro(self):
        """[summary]

        :return: A list of plotting directives
        :rtype: list
        """
        m = [self._projection, self._background]
        for r in self._layers:
            r.add_action(m)
        m += [
            self._rivers,
            self._borders,
            self._cities,
            self._foreground,
            self._grid,
            self._legend,
            self._title,
        ]
        return [x for x in m if x is not None]

    def wms_layers(self):

        self.finalise()

        m = [self._background]
        for r in self._layers:
            r.add_action(m)
        m += [
            self._rivers,
            self._borders,
            self._cities,
            self._foreground,
            self._grid,
        ]

        actions = [x for x in m if x is not None]
        return WMSLayer(actions, self._bounding_box, self._tmp)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from climetlab.utils import bytes_to_string
from climetlab.utils.bbox import BoundingBox


def test_bytes():
    assert bytes_to_string(10) == "10"
    assert bytes_to_string(1024) == "1 KiB"
    assert bytes_to_string(1024 * 1024) == "1 MiB"


def test_bbox():

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i, south=30, east=10 + i)
        assert bbox.width == 10, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i, south=30, east=350 + i)
        assert bbox.width == 350, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i, south=30, east=(10 + i) % 360)
        assert bbox.width == 10, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i, south=30, east=(350 + i) % 360)
        assert bbox.width == 350, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i % 360, south=30, east=10 + i)
        assert bbox.width == 10, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i % 360, south=30, east=350 + i)
        assert bbox.width == 350, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i % 360, south=30, east=(10 + i) % 360)
        assert bbox.width == 10, bbox

    for i in range(-365, 365):
        bbox = BoundingBox(north=90, west=i % 360, south=30, east=(350 + i) % 360)
        assert bbox.width == 350, bbox

    for i in range(-365, 365):
        bbox1 = BoundingBox(north=90, west=150 + i, south=30, east=170 + i)
        bbox2 = BoundingBox(north=90, west=-170 + i, south=30, east=-150 + i)
        bbox = bbox1.merge(bbox2)
        assert bbox.width == 60, (bbox1, bbox2, bbox)

        bbox = bbox2.merge(bbox1)
        assert bbox.width == 60, (bbox1, bbox2, bbox)

    with pytest.raises(ValueError):
        BoundingBox(north=-10, west=0, south=30, east=1)

    with pytest.raises(ValueError):
        BoundingBox(north=90, west=1, south=30, east=1)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import pytest

from climetlab import load_source, plot_map


def test_plot():
    for s in load_source("file", "docs/examples/test.grib"):
        plot_map(s)

        # test.grib fields endStep is 0, so datetime == valid_datetime
        assert s.datetime() == s.valid_datetime()

        # test shape
        assert s.shape == (11, 19)


@pytest.mark.skipif("GITHUB_WORKFLOW" in os.environ, reason="Not yet ready")
def test_sel():
    s = load_source("file", "docs/examples/test.grib")

    s.sel(shortName="2t")


@pytest.mark.skipif("GITHUB_WORKFLOW" in os.environ, reason="Not yet ready")
def test_multi():
    s1 = load_source(
        "cds",
        "reanalysis-era5-single-levels",
        product_type="reanalysis",
        param="2t",
        date="2021-03-01",
    )
    s2 = load_source(
        "cds",
        "reanalysis-era5-single-levels",
        product_type="reanalysis",
        param="2t",
        date="2021-03-02",
    )
    s3 = load_source(
        "cds",
        "reanalysis-era5-single-levels",
        product_type="reanalysis",
        param="2t",
        date=["2021-03-01", "2021-03-02"],
    )
    source = load_source("multi", s1, s2)
    for s in source:
        print(s)

    source.to_xarray()

    import xarray as xr

    print(s1)
    print(s1.path)
    print(s2.path)
    print(s3.path)

    print("--------------")
    s3.to_xarray()
    print("--------------")

    xr.open_mfdataset([s1.path, s2.path], engine="cfgrib")


if __name__ == "__main__":
    test_multi()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
from collections import defaultdict

from climetlab.core.data import get_data_entry

from . import magics_keys_to_actions
from .actions import lookup

LOG = logging.getLogger(__name__)


def _find_action(value, action):

    magics_keys = magics_keys_to_actions()

    # Guess the best action from the keys
    scores = defaultdict(int)
    special = 0
    for param in value.keys():

        if not param[0].isalpha():
            special += 1
            param = param[1:]

        acts = magics_keys.get(param, [])
        if len(acts) == 1:
            # Only consider unambiguous parameters
            scores[list(acts)[0]] += 1

    best = sorted((v, k) for k, v in scores.items())

    if len(best) == 0:
        LOG.warning("Cannot establish Magics action from [%r]", list(value.keys()))

    if len(best) >= 2 and best[0][0] == best[1][0]:
        LOG.warning(
            "Cannot establish Magics action from [%r], it could be %s or %s",
            list(value.keys()),
            best[0][1],
            best[1][1],
        )

    if len(best) > 0:
        action = lookup(best[0][1])

    return action, special


def _apply_dict(*, value, collection, action, default, target, options):

    if "update" in value:
        newvalue = {}
        for k, v in value.get("set", {}).items():
            if v is None:
                newvalue["-{}".format(k)] = v
            else:
                newvalue["+{}".format(k)] = v

        return apply(
            value=newvalue,
            collection=collection,
            action=action,
            default=default,
            target=target,
            options=options,
        )

    if "set" in value or "clear" in value:
        newvalue = {}
        for k, v in value.get("set", {}).items():
            newvalue["+{}".format(k)] = v

        for k in value.get("clear", []):
            newvalue["-{}".format(k)] = None

        return apply(
            value=newvalue,
            collection=collection,
            action=action,
            default=default,
            target=target,
            options=options,
        )

    if "+" in value or "-" in value:
        newvalue = {}
        for k, v in value.get("+", {}).items():
            newvalue["+{}".format(k)] = v

        for k in value.get("-", []):
            newvalue["-{}".format(k)] = None

        return apply(
            value=newvalue,
            collection=collection,
            action=action,
            default=default,
            target=target,
            options=options,
        )

    action, special = _find_action(value, action)
    if special:
        if special != len(value):
            raise ValueError(
                "Cannot set some attributes and override others %r" % list(value.keys())
            )

        result = target.update(action, value)
        if result is not None:
            return result

        raise ValueError(
            "Cannot override attributes %r (no matching style)" % list(value.keys())
        )

    return action(**value)


def _apply_true(*, value, collection, action, default, target, options):
    assert default is not True
    return apply(
        value=default,
        collection=collection,
        action=action,
        default=None,
        target=target,
        options=options,
    )


def _apply_string(*, value, collection, action, default, target, options):

    # TODO: Consider `value` being a URL (yaml or json)

    data = get_data_entry(collection, value).data

    options.update_if_not_set(**data.get("plot_map", {}))

    magics = data["magics"]
    actions = list(magics.keys())
    if len(actions) != 1:
        raise ValueError(
            "%s %s: one, and only one magics action can be defined in a yaml file: %r"
            % (collection, value, actions)
        )

    name = actions[0]
    action = lookup(name)
    kwargs = magics[name]
    return action(**kwargs)


def apply(*, value, collection=None, action=None, default=None, target, options):

    if value in (None, False):
        return None

    if value is True:
        return _apply_true(
            value=value,
            collection=collection,
            action=action,
            default=default,
            target=target,
            options=options,
        )

    if isinstance(value, dict):
        return _apply_dict(
            value=value,
            collection=collection,
            action=action,
            default=default,
            target=target,
            options=options,
        )

    if isinstance(value, str):
        return _apply_string(
            value=value,
            collection=collection,
            action=action,
            default=default,
            target=target,
            options=options,
        )

    raise ValueError("Unsupported type %s, %s (%s)" % (type(value), value, collection))



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import datetime

from climetlab.helpers import helper as get_helper


class NumpyArrayHelper:
    def __init__(self, data):
        self.data = data

    def plot_map(self, driver):
        # field = driver.option("field")
        # grid = driver.option("grid")
        metadata = get_helper(driver.option("metadata"))
        metadata = metadata.field_metadata()

        driver.bounding_box(
            north=metadata.get("north", 90),
            south=metadata.get("south", -90),
            west=metadata.get("west", 0),
            east=metadata.get("east", 360),
        )

        driver.plot_numpy(
            self.data.reshape(metadata.get("shape", self.data.shape)),
            metadata=metadata,
        )

    def to_datetime_list(self):
        return [datetime.datetime.fromtimestamp(x * 1e-9) for x in self.data.tolist()]


helper = NumpyArrayHelper



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.utils.conventions import normalise_string


class ParameterNormaliser:
    def __init__(self, convention=None):
        self.convention = convention

    def normalise(self, parameter):
        if isinstance(parameter, (list, tuple)):
            return [normalise_string(p, convention=self.convention) for p in parameter]
        else:
            return normalise_string(parameter, convention=self.convention)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import pandas as pd
import xarray as xr

from climetlab.core.metadata import annotate, annotation


class Owner:
    pass


def test_pandas_annotations():
    data = dict(a=["foo", "bar"], b=[1, 2])

    df1 = pd.DataFrame(data, columns=["a", "b"])

    obj1 = Owner()
    assert annotate(df1, obj1, foo=42) is df1
    assert "climetlab-0" in df1._metadata

    a1 = annotation(df1)

    assert a1.get("foo") == 42
    assert a1.owner is obj1

    df2 = df1[df1.b == 42]
    a2 = annotation(df2)
    assert a2.get("foo") == 42
    assert a2.owner is obj1

    assert a1 is a2

    del obj1

    assert a2.owner is None

    obj3 = Owner
    df3 = pd.DataFrame(data, columns=["a", "b"])
    annotate(df3, obj3, bar=42)

    a3 = annotation(df3)
    assert a1 is not a3

    assert "climetlab-0" in df3._metadata


def test_xarray_annotations():

    # Examples from xarray documentation

    # Data array
    ############

    data = np.random.rand(4, 3)
    locs = ["IA", "IL", "IN"]
    times = pd.date_range("2000-01-01", periods=4)
    xr1 = xr.DataArray(data, coords=[times, locs], dims=["time", "space"])
    obj1 = Owner()
    assert annotate(xr1, obj1, foo=42) is xr1

    a1 = annotation(xr1)
    assert a1.get("foo") == 42

    # Dataset
    #########

    temp = 15 + 8 * np.random.randn(2, 2, 3)
    precip = 10 * np.random.rand(2, 2, 3)
    lon = [[-99.83, -99.32], [-99.79, -99.23]]
    lat = [[42.25, 42.21], [42.63, 42.59]]

    xr2 = xr.Dataset(
        {
            "temperature": (["x", "y", "time"], temp),
            "precipitation": (["x", "y", "time"], precip),
        },
        coords={
            "lon": (["x", "y"], lon),
            "lat": (["x", "y"], lat),
            "time": pd.date_range("2014-09-06", periods=3),
            "reference_time": pd.Timestamp("2014-09-05"),
        },
    )

    annotate(xr2, obj1, bar=42)
    a1 = annotation(xr2)
    assert a1.get("bar") == 42

    # Dataset from Data array
    # annotation must be preserved

    # xr3 = xr1.to_dataset(name="test")
    # a3 = annotation(xr3)
    # assert a3.get("foo") == 42



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.datasets import Dataset
from climetlab.sources import DataSource

from .core.caching import CACHE as cache
from .core.metadata import init_metadata
from .core.settings import SETTINGS as settings
from .datasets import dataset, load_dataset
from .plotting import interactive_map, new_plot
from .plotting import options as plotting_options
from .plotting import plot_map
from .sources import load as load_source

# from climetlab.core.ipython import ipython_active


# import logging

__version__ = "0.3.3"


# if ipython_active:
#     logging.


__all__ = [
    "load_source",
    "load_dataset",
    "dataset",
    "plot_map",
    "interactive_map",
    "new_plot",
    "settings",
    "cache",
    "Dataset",
    "DataSource",
    "plotting_options",
]

init_metadata()



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.utils.bbox import BoundingBox, to_bounding_box

CONVERT = {
    list: lambda x: x.as_list(),
    tuple: lambda x: x.as_tuple(),
    dict: lambda x: x.as_dict(),
    BoundingBox: lambda x: x,
}


class BoundingBoxNormaliser:
    def __init__(self, format=BoundingBox):
        self.format = format

    def normalise(self, bbox):
        bbox = to_bounding_box(bbox)
        return CONVERT[self.format](bbox)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

from climetlab import load_source, source


def test_file_source_grib():
    s = load_source("file", "docs/examples/test.grib")
    assert len(s) == 2


def test_file_source_netcdf():
    s = load_source("file", "docs/examples/test.nc")
    assert len(s) == 2


@pytest.mark.skipif(sys.version_info < (3, 7), reason="Version 3.7 or greater needed")
def test_file_source_shortcut():
    s = source.file("docs/examples/test.grib")
    assert len(s) == 2


def test_file_source_mars():

    if not os.path.exists(os.path.expanduser("~/.ecmwfapirc")):
        pytest.skip("No ~/.ecmwfapirc")

    s = load_source(
        "mars",
        param=["2t", "msl"],
        levtype="sfc",
        area=[50, -50, 20, 50],
        grid=[1, 1],
        date="2012-12-13",
    )
    assert len(s) == 2


def test_file_source_cds_grib():

    if not os.path.exists(os.path.expanduser("~/.cdsapirc")):
        pytest.skip("No ~/.cdsapirc")

    s = load_source(
        "cds",
        "reanalysis-era5-single-levels",
        variable=["2t", "msl"],
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date="2012-12-12",
        time="12:00",
    )
    assert len(s) == 2


def test_file_source_cds_netcdf():

    if not os.path.exists(os.path.expanduser("~/.cdsapirc")):
        pytest.skip("No ~/.cdsapirc")

    s = load_source(
        "cds",
        "reanalysis-era5-single-levels",
        variable=["2t", "msl"],
        product_type="reanalysis",
        area=[50, -50, 20, 50],
        date="2012-12-12",
        time="12:00",
        format="netcdf",
    )
    assert len(s) == 2


def test_ulr_source_1():
    load_source("url", "http://download.ecmwf.int/test-data/metview/gallery/temp.bufr")


def test_ulr_source_2():
    load_source(
        "url", "https://github.com/ecmwf/climetlab/raw/master/docs/examples/test.grib"
    )


def test_ulr_source_3():
    load_source(
        "url", "https://github.com/ecmwf/climetlab/raw/master/docs/examples/test.nc"
    )


def zarr_not_installed():
    try:
        import s3fs
        import zarr

        return False
    except ImportError:
        return True


S3_URL = "https://storage.ecmwf.europeanweather.cloud/s2s-ai-competition/data/fixtures"


@pytest.mark.skipif(zarr_not_installed(), reason="Zarr or S3FS not installed")
def test_zarr_source_1():
    source = load_source(
        "zarr-s3",
        f"{S3_URL}/0.1.20/zarr/mini-rt-20200102.zarr",
    )
    ds = source.to_xarray()
    assert len(ds.forecast_time) == 1


@pytest.mark.skipif(zarr_not_installed(), reason="Zarr or S3FS not installed")
def test_zarr_source_2():
    import datetime

    from climetlab.utils.dates import to_datetime_list

    source = load_source(
        "zarr-s3",
        [
            f"{S3_URL}/0.1.20/zarr/mini-rt-20200109.zarr",
            f"{S3_URL}/0.1.20/zarr/mini-rt-20200102.zarr",
        ],
    )

    ds = source.to_xarray()
    assert len(ds.forecast_time) == 2

    dates = to_datetime_list(ds.forecast_time)
    assert dates[0] == datetime.datetime(2020, 1, 2)
    assert dates[1] == datetime.datetime(2020, 1, 9)

    dates = to_datetime_list(ds.forecast_time.values)
    assert dates[0] == datetime.datetime(2020, 1, 2)
    assert dates[1] == datetime.datetime(2020, 1, 9)


@pytest.mark.skipif(zarr_not_installed(), reason="Zarr or S3FS not installed")
def test_zarr_source_3():
    import datetime

    import numpy as np

    from climetlab.utils.dates import to_datetime_list

    source = load_source(
        "zarr-s3",
        [
            f"{S3_URL}/0.1.20/zarr/mini-hc-20200109.zarr",
            f"{S3_URL}/0.1.20/zarr/mini-hc-20200102.zarr",
        ],
    )
    ds = source.to_xarray()
    assert len(ds.forecast_time) == 8

    dates = to_datetime_list(ds.forecast_time)
    assert dates[0] == datetime.datetime(2000, 1, 2)
    assert dates[1] == datetime.datetime(2000, 1, 9)
    assert dates[2] == datetime.datetime(2001, 1, 2)
    assert dates[3] == datetime.datetime(2001, 1, 9)

    dates = to_datetime_list(ds.forecast_time.values)
    assert dates[0] == datetime.datetime(2000, 1, 2)
    assert dates[1] == datetime.datetime(2000, 1, 9)
    assert dates[2] == datetime.datetime(2001, 1, 2)
    assert dates[3] == datetime.datetime(2001, 1, 9)


if __name__ == "__main__":
    test_zarr_source_2()



import logging
import os
import socket
import sys
import time
import traceback
import uuid
from contextlib import closing

import jinja2
import requests
import skinnywms
from flask import Flask, Response, jsonify, render_template, request, send_file
from IPython.lib import backgroundjobs as bg
from skinnywms.datatypes import Availability, DataLayer, Field
from skinnywms.plot.magics import Plotter, Styler
from skinnywms.server import WMSServer

from climetlab import new_plot
from climetlab.core.ipython import display

jobs = bg.BackgroundJobManager()

application = Flask(__name__)

# Be quiet
logging.getLogger("werkzeug").setLevel(logging.CRITICAL)

loader = jinja2.ChoiceLoader(
    [
        application.jinja_loader,
        jinja2.FileSystemLoader(
            os.path.join(os.path.dirname(skinnywms.__file__), "templates")
        ),
    ]
)

application.jinja_loader = loader


class CliMetLabField(Field):
    def __init__(self, obj):
        self.obj = obj
        self._plot = new_plot().plot_map(self.obj).wms_layers()

    @property
    def name(self):
        try:
            return self.obj.name
        except Exception:
            return "?"

    @property
    def title(self):
        try:
            return self.obj.title
        except Exception:
            return "?"

    @property
    def time(self):
        return None

    @property
    def styles(self):
        return ["default"]

    def render(self, context, driver, style):
        return self._plot.actions

    def bounding_box(self):
        return self._plot.bounding_box


class CliMetLabLayer(DataLayer):
    def __init__(self, obj):
        super().__init__(CliMetLabField(obj))

    def bounding_box(self):
        return self._first.bounding_box()


class CliMetLabAvailability(Availability):
    def __init__(self, obj):
        super().__init__()
        self._layers["climetlab"] = CliMetLabLayer(obj)

    def bounding_box(self):
        return self._layers["climetlab"].bounding_box()


class CliMetLabWMSServer(WMSServer):
    def process(self, *args, **kwargs):
        try:
            return super().process(*args, **kwargs)
        except Exception:
            print(traceback.format_exc())
            raise


@application.route("/wms/<uid>", methods=["GET"])
def wms(uid):

    try:

        svr = SERVERS[uid]

        reply = svr.process(
            request,
            Response=Response,
            send_file=send_file,
            render_template=render_template,
            reraise=True,
        )

        return reply
    except Exception:
        print(traceback.format_exc())
        return Response(traceback.format_exc(), mimetype="text/plain", status=500)


@application.route("/status", methods=["GET"])
def status():
    return jsonify(
        dict(
            pid=os.getpid(),
        )
    )


def _task(port):
    # print("Start task", port)
    os.environ["WERKZEUG_RUN_MAIN"] = "true"  # Get rid of any banner
    try:
        application.run(
            host="localhost", port=port, load_dotenv=False, debug=False, threaded=4
        )
    except Exception as e:
        print("WMS server crashed:", e)
    # print("End task", port)


class State:
    job = None
    url = None


STATE = State()


def start_wms():

    if STATE.job is not None and STATE.url is not None:
        if STATE.job.status == STATE.job.stat_running:
            return STATE.url
        jobs.flush()
        STATE.job = STATE.url = None

    for _ in range(10):

        with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
            s.bind(("localhost", 0))
            port = s.getsockname()[1]

        STATE.job = jobs.new(_task, port)
        while STATE.job.status is STATE.job.stat_created:
            time.sleep(0.1)

        n = 0
        max_tries = 5
        while n < max_tries:
            status = f"http://localhost:{port}/status"
            try:
                r = requests.get(status)
                r.raise_for_status()
                if r.json()["pid"] == os.getpid():
                    STATE.url = f"http://localhost:{port}/wms"
                    return STATE.url
                else:
                    break
            except requests.exceptions.HTTPError as e:
                print("WMS status at", status, e)
                break
            except requests.exceptions.ConnectionError as e:
                if n + 1 == max_tries:
                    print("WMS status at", status, e)
                time.sleep(1)
                n += 1

        time.sleep(1)

    raise Exception("Cannot start WMS server")


SERVERS = {}


# def gone(*args, **kwargs):
#     print('gone', args, kwargs)


def interactive_map(obj, **kwargs):
    # Prefer `folium` to `ipyleafet` as it does not
    # rely on ipywidgets, that are not always available
    # from ipyleaflet import Map, WMSLayer, projections, FullScreenControl
    import folium
    import folium.plugins
    from branca.element import MacroElement
    from jinja2 import Template

    class NoScrollZoom(MacroElement):
        _name = "NoScrollZoom"
        _template = Template(
            """
            {% macro header(this,kwargs) %}
            {% endmacro %}
            {% macro html(this,kwargs) %}
            {% endmacro %}
            {% macro script(this,kwargs) %}
            {{ this._parent.get_name() }}.scrollWheelZoom.disable();
            {% endmacro %}
        """
        )

    uid = str(uuid.uuid1())
    # TODO: use weak ref

    availability = CliMetLabAvailability(obj)
    SERVERS[uid] = CliMetLabWMSServer(availability, Plotter(), Styler())
    url = "{}/{}".format(start_wms(), uid)

    bbox = availability.bounding_box()
    center = (0, 0)
    zoom = 1

    if bbox is not None:
        center = (bbox.north + bbox.south) / 2, (bbox.east + bbox.west) / 2
        zoom = 1 / max((bbox.north - bbox.south) / 180, (bbox.east - bbox.west) / 360)
        # print("zoom", zoom)
        zoom = (2 * zoom + 88) / 27
        # print("zoom", zoom)

    # 10 => 4
    # 37 => 6

    m = folium.Map(
        zoom_start=zoom, location=center
    )  # , crs=projections.Base)  # basemap={},

    folium.raster_layers.WmsTileLayer(
        url=url, layers=["climetlab"], transparent=True, fmt="image/png", **kwargs
    ).add_to(m)

    # https://github.com/python-visualization/folium/blob/master/examples/Plugins.ipynb

    folium.plugins.Fullscreen(force_separate_button=True).add_to(m)
    NoScrollZoom().add_to(m)

    if bbox is not None:
        m.fit_bounds([[bbox.south, bbox.east], [bbox.north, bbox.west]])

    # import weakref
    # availability.map = weakref.ref(m, gone)

    # return m
    return display(m)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from climetlab import load_source

from . import Meteonet


class MeteonetWeatherModels(Meteonet):
    """
    See https://github.com/meteofrance/meteonet
    """

    def __init__(self):
        pass

    def _load(
        self, model="arome", variable="2m", domain="NW", date="20180501", time="0000"
    ):

        url = "{url}/weather_models/{model}_{variable}_{domain}_{date}{time}00.grib".format(
            url=self.URL,
            variable=variable,
            model=model,
            domain=domain,
            date=date,
            time=time,
        )
        self.source = load_source("url", url)


dataset = MeteonetWeatherModels



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import re

import numpy as np

# datetime.fromisoformat() only available from Python3.7
# from backports.datetime_fromisoformat import MonkeyPatch
from dateutil.parser import isoparse, parse

# from collections import defaultdict
from climetlab.helpers import helper

# MonkeyPatch.patch_fromisoformat()


VALID_DATE = re.compile(r"\d\d\d\d-?\d\d-?\d\d([T\s]\d\d:\d\d(:\d\d)?)?Z?")


def parse_date(dt):

    if not isinstance(dt, str):
        return to_datetime(dt)

    if not VALID_DATE.match(dt):
        raise ValueError(f"Invalid datetime '{dt}'")

    try:
        return datetime.datetime.fromisoformat(dt)
    except Exception:
        pass

    try:
        return isoparse(dt)
    except ValueError:
        pass

    return parse(dt)


def to_datetime(dt):
    if isinstance(dt, datetime.datetime):
        return dt

    if isinstance(dt, datetime.date):
        return datetime.datetime(dt.year, dt.month, dt.day)

    if isinstance(dt, np.datetime64):
        # Looks like numpy dates conversion vary
        dt = dt.astype(datetime.datetime)

        if isinstance(dt, datetime.datetime):
            return dt

        if isinstance(dt, datetime.date):
            return to_datetime(dt)

        if isinstance(dt, int):
            return datetime.datetime.utcfromtimestamp(dt * 1e-9)

        raise ValueError("Failed to convert numpy datetime {}".format((dt, type(dt))))

    if isinstance(dt, str):
        return parse_date(dt)

    if getattr(dt, "to_datetime", None) is None:
        dt = helper(dt)

    return to_datetime(dt.to_datetime())


def _mars_list(start, end, by):
    assert by > 0, by
    assert end >= start
    result = []
    while start <= end:
        result.append(start)
        start = start + datetime.timedelta(days=by)
    return result


def to_datetime_list(datetimes):

    if isinstance(datetimes, str):
        # MARS style lists
        bits = datetimes.split("/")
        if len(bits) == 3 and bits[1].lower() == "to":
            return _mars_list(to_datetime(bits[0]), to_datetime(bits[2]), 1)

        if len(bits) == 5 and bits[1].lower() == "to" and bits[3].lower() == "by":
            return _mars_list(to_datetime(bits[0]), to_datetime(bits[2]), int(bits[4]))

        try:
            return to_datetime_list(bits)
        except Exception:
            pass

    if isinstance(datetimes, (datetime.datetime, np.datetime64, str)):
        return to_datetime_list([datetimes])

    if isinstance(datetimes, (list, tuple)):
        if len(datetimes) == 3 and datetimes[1].lower() == "to":
            return _mars_list(to_datetime(datetimes[0]), to_datetime(datetimes[2]), 1)

        if (
            len(datetimes) == 5
            and datetimes[1].lower() == "to"
            and datetimes[3].lower() == "by"
        ):
            return _mars_list(
                to_datetime(datetimes[0]), to_datetime(datetimes[2]), int(datetimes[4])
            )

        return [to_datetime(x) for x in datetimes]

    if getattr(datetimes, "to_datetime_list", None) is None:
        datetimes = helper(datetimes)

    return to_datetime_list(datetimes.to_datetime_list())


def to_dates_and_times(datetimes_list):
    assert False, datetimes_list
    # result = []
    # datetimes = defaultdict(set)

    # for dt in to_datetime_list(datetimes_list):
    #     datetimes[dt.date())].add(dt.time())

    # timedates = defaultdict(set)
    # for date, times in sorted(datetimes.items()):
    #     times = tuple(sorted(times))
    #     timedates[times].add(date)

    # for times, dates in timedates.items():
    #     result.append((tuple(sorted(dates)), times))

    # return sorted(result)


def to_date_list(obj):
    return sorted(set(to_datetime_list(obj)))



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import weakref


class Reader:
    def __init__(self, source, path):
        self._source = weakref.ref(source)
        self.path = path

    @property
    def source(self):
        return self._source()

    def mutate(self):
        # Give a chance to `directory` or `zip` to change the reader
        return self

    def sel(self, *args, **kwargs):
        raise NotImplementedError()


def grib_reader(source, path):
    from .grib import GRIBReader

    return GRIBReader(source, path)


def bufr_reader(source, path):
    from .bufr import BUFRReader

    return BUFRReader(source, path)


def netcdf_reader(source, path):
    from .netcdf import NetCDFReader

    return NetCDFReader(source, path)


def odb_reader(source, path):
    from .odb import ODBReader

    return ODBReader(source, path)


def csv_reader(source, path):
    from .csv import CSVReader

    return CSVReader(source, path)


def zip_reader(source, path):
    from .zip import ZIPReader

    return ZIPReader(source, path)


def directory_reader(source, path):
    from .directory import DirectoryReader

    return DirectoryReader(source, path)


READERS = {
    b"GRIB": grib_reader,
    b"BUFR": bufr_reader,
    b"\x89HDF": netcdf_reader,
    b"CDF\x01": netcdf_reader,
    b"CDF\x02": netcdf_reader,
    b"\xff\xffOD": odb_reader,
    b"PK\x03\x04": zip_reader,
    ".csv": csv_reader,
    ".d": directory_reader,
}


def reader(source, path):

    _, extension = os.path.splitext(path)
    if extension in READERS:
        return READERS[extension](source, path).mutate()

    if os.path.isdir(path):
        return directory_reader(source, path).mutate()

    with open(path, "rb") as f:
        header = f.read(4)

    if header in READERS:
        return READERS[header](source, path).mutate()

    raise ValueError(
        "Unsupported file {} (header={}, extension={})".format(path, header, extension)
    )
